<footer>
  {% if (theme_prev_next_buttons_location == 'bottom' or theme_prev_next_buttons_location == 'both') and (next or prev) %}
    <div class="rst-footer-buttons" role="navigation" aria-label="footer navigation">
      {% if next %}
        <a href="{{ next.link|e }}" class="btn btn-neutral float-right" title="{{ next.title|striptags|e }}" accesskey="n" rel="next">{{ _('Next') }} <span class="fa fa-arrow-circle-right"></span></a>
      {% endif %}
      {% if prev %}
        <a href="{{ prev.link|e }}" class="btn btn-neutral float-left" title="{{ prev.title|striptags|e }}" accesskey="p" rel="prev"><span class="fa fa-arrow-circle-left"></span> {{ _('Previous') }}</a>
      {% endif %}
    </div>
  {% endif %}

  <hr/>

  <div role="contentinfo">

    <p>
    {%- if show_copyright %}
      {%- if hasdoc('copyright') %}
        {% trans path=pathto('copyright'), copyright=copyright|e %}&copy; <a href="{{ path }}">Copyright</a> {{ copyright }}{% endtrans %}
      {%- else %}
        {% trans copyright=copyright|e %}&copy; Copyright {{ copyright }}{% endtrans %}
      {%- endif %}
    {%- endif %}

    {%- if build_id and build_url %}
      {% trans build_url=build_url, build_id=build_id %}
        <span class="build">
          Build
          <a href="{{ build_url }}">{{ build_id }}</a>.
        </span>
      {% endtrans %}
    {%- elif commit %}
      {% trans commit=commit %}
        <span class="commit">
          Revision <code>{{ commit }}</code>.
        </span>
      {% endtrans %}
    {%- elif last_updated %}
      <span class="lastupdated">
        {% trans last_updated=last_updated|e %}Last updated on {{ last_updated }}.{% endtrans %}
      </span>
    {%- endif %}

    </p>

  </div>
  {%- if show_sphinx %}
  {% trans %}
  <p>
  Built with <a href="http://sphinx-doc.org/">Sphinx</a> using a <a href="https://github.com/rtfd/sphinx_rtd_theme">theme</a> provided by <a href="https://readthedocs.org">Read the Docs</a>{% endtrans %}.
  </p>
  {%- endif %}
  

  <p style="text-align:center;">
    <a class="affiliation" href="https://www.psu.edu/" rel="noopener" target="_blank">
      <img src="{{pathto('_static/PennStateLogo.jpeg', 1)}}" alt="Pennsylvania State University" style="width: 3vw;"  hspace="20">
    </a>
    <a class="affiliation" href="https://www.gssi.it/" rel="noopener" target="_blank">
      <img src="{{pathto('_static/GSSI_logo.jpeg', 1)}}" alt="GSSI" style="width: 6vw;"  hspace="20">
    </a>
    <a class="affiliation" href="https://www.uchicago.edu/" rel="noopener" target="_blank">
      <img src="{{pathto('_static/UChicago_logo.png', 1)}}" alt="University of Chicago" style="width: 2vw;" hspace="20">
    </a>
  
  </p>
  {%- block extrafooter %}
  
  {% endblock %}
</footer>

